{
--
-- Parser for extracting element names with corresponding features 
-- out of guidsl model files
--
-- Author: Wolfgang Scholz, 2010-06
--
module GuiDSL where
import Char
import ProductLine


}

%name parseGuiDSL
-- [Token] -> ([DProduction], [DConstr])
%tokentype { Token }
%error { parseError }

%token
      '%%'            { TokenEndFM }
      ':'             { TokenColon }
      ';'             { TokenSemicolon }
      '::'            { TokenDColon }
      '|'             { TokenDash }
      '+'             { TokenPlus }
      '*'             { TokenStar }
      '['             { TokenOB }
      ']'             { TokenCB }
      '('             { TokenORB }
      ')'             { TokenCRB }
      and             { TokenAnd }
      or              { TokenOr }
      not             { TokenNot }
      iff             { TokenIff }
      implies         { TokenImplies }
      var             { TokenFeat $$ }

-- operator precedence for constraints (implies is intended to occur only once)
%left '('
%left not
%left and
%left or
%left iff
%%

ProductLine :: { ([DProduction], [DConstr]) }
    : FeatureModel '%%' Constraints { ($1,$3) }
    | FeatureModel                  { ($1,[]) }

-- : var [] :: * +
-- a child is selected -> the parent is selected
-- the parent of an alternative structure is selected -> exactly one child is selected
-- the parent of an 'and' structure is selected -> all children are selected (-optional)
-- the parent of an 'or' structure is selected -> at least one child is selected 
-- the parent of a mandatory feature is selected -> mandatory feature is selected 

FeatureModel :: { [DProduction] }
    : Production FeatureModel { $1 : $2 }
    | Production              { [$1] }

Production :: { DProduction }
    : var ':' Patterns        { DProd $1 $3 }

Patterns :: { DPatterns }
    : Pattern '|' Patterns    { DPatns $1 $3 }
    | Pattern ';'             { DPatn $1 }

Pattern :: { DPattern }
    : TermSeries '::' var     { DSeriesPat $3 $1 } --name of pattern first here
    | OptionalTerm            { DSinglePat $1 }

TermSeries :: { DTermSeries }
    : Term TermSeries         { DTrmS $1 $2 }
    | Term                    { DTrmSsingle $1 }

Term :: { DTerm } --conjunction term
    : var                     { DTrm $1 }
    | var '+'                 { DTrmPlus $1 }
    | '[' var ']'             { DTrmBrackets $2 } 
    | var '*'                 { DTrmStar $1 }

OptionalTerm :: { DOptionalTerm }
    : var                     { DOptTerm $1 }

-- Constraints
Constraints :: { [DConstr] }
    : Constraint ';' Constraints { $1 : $3 }
    | Constraint ';'             { [$1] }

Constraint :: { DConstr }
    : Exp implies Exp        { DConstrImpl $1 $3 }
    | Exp iff Exp            { DConstrIff $1 $3 }
    -- not implemented: let (new variables)

Exp :: { DConstrExp }
    : var                     { F $1 }
    | '(' Exp ')'             { $2 }
    | not Exp                 { ExpNot $2 }
    | Exp and Exp             { ExpAnd $1 $3 }
    | Exp or Exp              { ExpOr $1 $3 }
    --note: operator precedence not tested
{

data DProduction 
      = DProd String DPatterns
      deriving Show

data DPatterns
      = DPatns DPattern DPatterns
      | DPatn DPattern 
      deriving Show

data DPattern -- a pattern is a term series followed by a name / a single var
      = DSeriesPat String DTermSeries --beware: name of pattern first here ! 
      | DSinglePat DOptionalTerm
      deriving Show

data DTermSeries
      = DTrmS DTerm DTermSeries
      | DTrmSsingle DTerm
      deriving Show

data DTerm = DTrm String | DTrmPlus String | DTrmStar String | DTrmBrackets String
      deriving Show
data DOptionalTerm = DOptTerm String
      deriving Show

data DConstr = DConstrImpl DConstrExp DConstrExp
             | DConstrIff DConstrExp DConstrExp deriving Show
data DConstrExp = F String
                | ExpNot DConstrExp
                | ExpAnd DConstrExp DConstrExp | ExpOr DConstrExp DConstrExp deriving Show

data Token
      = TokenEndFM
      | TokenColon
      | TokenDColon
      | TokenSemicolon
      | TokenDash
      | TokenPlus
      | TokenStar
      | TokenOB
      | TokenCB
      | TokenORB
      | TokenCRB
      | TokenAnd
      | TokenOr
      | TokenNot
      | TokenIff
      | TokenImplies
      | TokenFeat String

instance Show Token where
    show TokenEndFM = "EOF"
    show TokenColon = ":"
    show TokenDColon = "::"--
-- Parser for extracting element names with corresponding features 
-- out of *.features.rsf files
--
-- Author: Wolfgang Scholz, 2010-06
--
    show TokenSemicolon = ";"
    show TokenDash = "-"
    show TokenPlus = "+"
    show TokenStar = "*"
    show TokenOB = "["
    show TokenCB = "]"
    show TokenORB = "("
    show TokenCRB = ")"
    show TokenAnd = "and"
    show TokenOr = "or"
    show TokenNot = "not"
    show TokenIff = "iff"
    show TokenImplies = "implies"
    show (TokenFeat str) = str


parseError :: [Token] -> a
parseError [] = error "Parse error"
parseError (x:xs) = error ("Parse error before : \"" 
    ++ (foldr (++) "" (map (' ':) (map show (take 12 (x:xs))))) ++ "\"")


-- lexer
lexer :: String -> [Token]
lexer [] = []
lexer ('%':('%':cs)) = TokenEndFM : lexer cs 
lexer (':':(':':cs)) = TokenDColon : lexer cs 
lexer (':':cs) = TokenColon : lexer cs
lexer (';':cs) = TokenSemicolon : lexer cs
lexer ('|':cs) = TokenDash : lexer cs
lexer ('+':cs) = TokenPlus : lexer cs
lexer ('*':cs) = TokenStar : lexer cs
lexer ('[':cs) = TokenOB : lexer cs
lexer (']':cs) = TokenCB : lexer cs
lexer ('(':cs) = TokenORB : lexer cs
lexer (')':cs) = TokenCRB : lexer cs
lexer (c:cs) 
      | isSpace c = lexer cs
      | otherwise = lexVar (c:cs)

lexVar cs =
   case span (\x -> (isAlpha x) || (isDigit x) || (x=='_')) cs of
      ("and",rest) -> TokenAnd : lexer rest
      ("or",rest) -> TokenOr : lexer rest
      ("iff",rest) -> TokenIff : lexer rest
      ("not",rest)  -> TokenNot : lexer rest
      ("implies",rest) -> TokenImplies : lexer rest
      (var,rest)   -> TokenFeat var : lexer rest

-- discards comments
removeComments :: String -> String
removeComments [] = ""
removeComments ('/':('/':xs)) = discardLine xs
    where discardLine :: String -> String --until EOL
          discardLine [] = ""
          discardLine (x:xs) | isPrint x = discardLine xs
                             | otherwise = removeComments xs
removeComments ('/':('*':xs)) = discardComment xs
    where discardComment :: String -> String --until "*/"
          discardComment [] = ""
          discardComment ('*':('/':xs)) = removeComments xs
          discardComment (x:xs) = discardComment xs
removeComments (x:xs) = x : (removeComments xs)

-- discards unneeded formatting data in GUIDSL model
discardFormatting :: String -> String
discardFormatting [] = []
discardFormatting ('#':('#':xs)) = [] --TODO: change back to '##' - constraints
discardFormatting (x:xs) = x : (discardFormatting xs)

main = getContents >>= print . parseGuiDSL . lexer . discardFormatting . removeComments 

}

--happy example.y 
--happy example.y -i 



