module ProductLine (ProductLine, FeatureModel, DomainConstraint(..), FeatureConstraint(..), 
checkFeatureModelFullyAbstract, checkReferenceFullyAbstract, getUnsuppliedReferences) where 

-- Usage of many overloaded methods will weaken the analysis.
-- As input data does not contain types, momentarily methods with same name
-- are ambiguous. If a method causes problems, all instances of it must 
-- cause the problem. If a method solves problems, any instance of it solves the 
-- problem.
-- To undo this behavior, visit these locations:
-- - checkReferenceFullyAbstract
-- - generateRefFMtree
-- - getUnsuppliedReferences

import FeatureStructure
import Funsat.Types
import Funsat.Circuit hiding (not)
import Funsat.Solver
import Funsat.Resolution
import Data.Set hiding (union)
import qualified Data.Set(map,union)
import qualified Prelude(map)
import Prelude
import Maybe
import Data.List
import qualified Data.Map

type ProductLine = (FeatureSet, FeatureModel)
type FeatureModel = [DomainConstraint] --numbers are references into a product line's feature set

emptyModel :: FeatureModel
emptyModel = []
--emptyModel = (CNF {numVars=0,numClauses=0,clauses=Data.Set.empty})

-- generate referencial feature model proposition tree for given FeatureSet (references are in features)
generateReferencialFeatureModelTree :: FeatureSet -> Tree Feature
generateReferencialFeatureModelTree fs = 
    foldr TAnd TTrue (Prelude.map (generateRefFMtree fs) (toList fs))

-- generate feature model prop tree out of given references
-- fs = all features, f = feature which houses reference sources
generateRefFMtree :: FeatureSet -> Feature -> Tree Feature
generateRefFMtree fs f = 
    foldr TAnd TTrue refFM
    where
        -- refFM = union of all references' generated referencial feature model
        refFM :: [Tree Feature]
        --refFM = Prelude.map (generateRef1DependencyTree fs f) (toList (references f))
        refFM = Prelude.map (generateRefSimpleDependencyTree fs) (toList (references f))
-- generates referencial feature model proposition tree out of one reference
generateRef1DependencyTree :: FeatureSet -> Feature -> Reference -> Tree Feature
generateRef1DependencyTree fs f (s,t)
    = if (Data.Set.null (Data.Set.filter (==s)  (elements f))) -- <-- is an assertion
    then error ("Feature '"++(show f)++"' misses (typed) source to own reference '(("
         ++(show(elName s))++","++(show(elType s))++")," ++(show t) ++ ")'. "++
         (let es = [e | e <- (toList(elements f)), elName e == elName s] in
             if ([] /= es) then ("The existing element is "
                 ++(show(elName(head es)))++","++(show(elType(head es)))++"))")
             else ("Elements are: "++(show (Data.Set.toList(elements f))))))
    else createFeatureDependencyTree (toList fs) f (getFeaturesWithElement fs t)
    where
    -- create a dependency clause from #2 to any of #3 (numbers obtained from #1)
    -- fs = all features, f = source of dependence, ts = targets of dependence
    createFeatureDependencyTree :: [Feature] -> Feature -> FeatureSet -> Tree Feature
    createFeatureDependencyTree fs f ts --create dep to every feature that holds the target of ref 
        | (member f ts) = TTrue --self reference: no restrictions
        | Data.Set.null ts = TNot (TLeaf f) -- no providing feature: never admittable
        | otherwise = TOr (TNot (TLeaf f)) (foldr1 TOr (Prelude.map TLeaf (toList ts)))

-- like generateRef1DependencyTree, but reference sources are 
generateRefSimpleDependencyTree :: FeatureSet -> Reference -> Tree Feature
generateRefSimpleDependencyTree fs (s,t) =
    if (length ts == 0) then TTrue
    else if (length ss == 0) then (foldr TOr TFalse (Prelude.map TLeaf ts))
         else TOr (TNot (foldr TAnd TTrue (Prelude.map TLeaf ss))) (foldr TOr TFalse (Prelude.map TLeaf ts))
    where ss = toList $ getFeaturesWithElement fs s
          ts = toList $ getFeaturesWithElement fs t

propTreeToSolverInput :: (Ord a, Show a) => Tree a -> CircuitProblem a
propTreeToSolverInput tree = toCNF (runShared (castCircuit tree))

-- try: problemCnf (propTreeToSolverInput _)
-- try: solve1 (problemCnf (propTreeToSolverInput _))

-- runs funsat to solve the given proposition tree 
solveTree :: (Ord a, Show a) => Tree a -> (Solution, Stats, Maybe ResolutionTrace)
solveTree tree = solve1 (problemCnf (propTreeToSolverInput tree))

-- gives a readable solution to a given tree, but only if a solution exists
projectTreeSolution :: (Ord a, Show a) => Tree a -> BEnv a
projectTreeSolution tree = 
    projectCircuitSolution (firstOf3 (solve1 (problemCnf problem))) problem
    where problem = propTreeToSolverInput tree
firstOf3 :: (a,b,c) -> a
firstOf3 (x,_,_) = x


-- build all configurations allowed by feature model
-- warning: running time is exponential
-- momentarily, no superimposition is implemented
buildAllConfigurations :: ProductLine -> Set [Feature]
buildAllConfigurations (fs,fm) = exhaustiveAllSolutions 
    (toList fs) (domainConstraintsToPropTree fm)

-- build all configurations allowed by references
-- warning: running time is exponential
-- momentarily, no superimposition is implemented
buildAllRefConfigurations :: ProductLine -> Set [Feature]
buildAllRefConfigurations (fs,_) = exhaustiveAllSolutions 
    (toList fs) (generateReferencialFeatureModelTree fs)

-- exhaustiveAllSolutions
-- running time: 2 ^ length(#1)
-- #1: features used in the tree, #2: feature model, #R: all valid configurations
exhaustiveAllSolutions :: [Feature] -> Tree Feature -> Set [Feature]
exhaustiveAllSolutions fs fm = fromList
    [buildConfiguration fs c | c <- combinations (length fs), checkCombination fs c fm]
    where buildConfiguration :: [Feature] -> [Bool] -> [Feature]
          buildConfiguration [] [] = []
          buildConfiguration [] _ = error "buildConfiguration - no features selectable" 
          buildConfiguration _ [] = error "buildConfiguration - no selection" 
          buildConfiguration (f:fs) (True:cs) = f : (buildConfiguration fs cs)
          buildConfiguration (f:fs) (False:cs) = buildConfiguration fs cs

-- constructs the Bool power set of #1 elements 
combinations :: Int -> [[Bool]]
combinations 1 = [[False],[True]]
combinations n = [ t:c |  t <- [False,True], c <- combinations (n-1)]

-- checks if given Feature combination is valid for the given feature model
-- running time: unoptimized polynomial (to the number of nodes in #3)
-- #1: features used in the tree, #2: configuration to test, #3 feature model 
checkCombination :: [Feature] -> [Bool] -> Tree Feature -> Bool
checkCombination fs cfg TTrue = True
checkCombination fs cfg TFalse = False
checkCombination fs cfg (TLeaf v) = cfg !! (fromJust (elemIndex v fs))
checkCombination fs cfg (TNot t) = Prelude.not (checkCombination fs cfg t)
checkCombination fs cfg (TAnd t1 t2) = (checkCombination fs cfg t1) && (checkCombination fs cfg t2)
checkCombination fs cfg (TOr t1 t2) = (checkCombination fs cfg t1) || (checkCombination fs cfg t2)
checkCombination fs cfg (TXor t1 t2) = (chkt1 && chkt2) || ((not chkt1) && (not chkt2))
    where chkt1 = checkCombination fs cfg t1
          chkt2 = checkCombination fs cfg t2
checkCombination fs cfg _ = error "TIff, TOnlyIf, TIte not implemented"


-- checks if the given domain constraints imply the feature model generated by references
-- must be universally quantified (for all allowed feature configurations)
-- The negation is tested, as the solver only checks satisfiability
-- The solution must result in: unsatisfiablility
checkFeatureModelFullyAbstractRaw :: ProductLine -> (Solution, Stats, Maybe ResolutionTrace)
checkFeatureModelFullyAbstractRaw (fs,dcs) = solveTree (TNot 
    (TOr (TNot (domainConstraintsToPropTree dcs)) 
    (generateReferencialFeatureModelTree fs)))

projectFeatureModelAbstractionViolation :: ProductLine -> BEnv Feature
projectFeatureModelAbstractionViolation (fs,dcs) = projectTreeSolution 
    (TNot (TOr (TNot (domainConstraintsToPropTree dcs))
    (generateReferencialFeatureModelTree fs)))

-- checks if the feature model has solutions (if yes, gives an example)
checkFeatureModelSatisfiable :: ProductLine -> (Bool, Maybe [Feature])
checkFeatureModelSatisfiable (fs,dcs) = hlp (solve1 (problemCnf solverInput))
    where 
        solverInput = propTreeToSolverInput (domainConstraintsToPropTree dcs)
        hlp (Sat ass, stats, Nothing) = 
            (True, Just (hlp2 (Data.Map.toList (projectCircuitSolution (Sat ass) solverInput))))
        hlp (Unsat ass, stats, Just resolutionTrace) = 
            (False, Nothing)--Right (genUnsatCore resolutionTrace))
        hlp2 :: [(Feature, Bool)] -> [Feature]
        hlp2 [] = []
        hlp2 ((f,False):fs) = hlp2 fs
        hlp2 ((f,True):fs) = f:(hlp2 fs)


-- A note to full abstractness and maximality checks:
-- Solver input is exactly the opposite of what I want to check.
-- I don't need unsatisfiable minimal cores (as the solver would give on counterexamples)
-- but I need full counterexamples (solutions to the opposite of what I want to find out)

-- checks if the given feature model implies the feature model constructed by references
-- result: (True,Nothing) if feature model is fully abstract
--         (False, Just [Feature]) otherwise; the list contains only feature names
checkFeatureModelFullyAbstract :: ProductLine -> (Bool, Maybe [Feature]) --(BEnv Feature)
checkFeatureModelFullyAbstract (fs,dcs) = hlp (solve1 (problemCnf solverInput))
    where 
        solverInput = propTreeToSolverInput (TNot (TOr (TNot (domainConstraintsToPropTree dcs)) 
                      (generateReferencialFeatureModelTree fs)))
        hlp (Sat ass, stats, Nothing) = 
            (False, Just (enrichFeatures fs 
                (hlp2 (Data.Map.toList (projectCircuitSolution (Sat ass) solverInput)))))
        hlp (Unsat ass, stats, Just resolutionTrace) = 
            (True, Nothing)--Right (genUnsatCore resolutionTrace))
        hlp2 :: [(Feature, Bool)] -> [Feature]
        hlp2 [] = []
        hlp2 ((f,False):fs) = hlp2 fs
        hlp2 ((f,True):fs) = f:(hlp2 fs)
-- adds information from #1 to each feature in list #2 (like elements/references)
-- $2: list of features with names (result will have elems/refs from corresponding $1)
-- running time: O(f * l)  [f=#features in #1, l=#features in #2]
enrichFeatures :: Set Feature -> [Feature] -> [Feature]
enrichFeatures _ [] = []
enrichFeatures fs (n:ns) = if (not (Data.Set.member n fs)) 
            then error ("product line inconsistency: referenced feature '"++(show n)
                 ++"' is not in feature list: "++(show fs))
            else (Data.Set.findMin (Data.Set.intersection fs (Data.Set.singleton n))) : (enrichFeatures fs ns)

-- checks exhaustively, if the given feature model implies the feature model constructed by references
-- result: (True, empty) if feature model is fully abstract
--         (False, fromList [fSet1, fSet2, ..]) otherwise, list contains all counterexamples
checkFeatureModelFullyAbstractExhaustive :: ProductLine -> (Bool, Set [Feature])
checkFeatureModelFullyAbstractExhaustive pl 
    | Data.Set.size (result) > 0 = (False,result)
    | otherwise                  = (True, result)
    where result = difference (buildAllConfigurations pl) (buildAllRefConfigurations pl)

-- checks if the given feature model implies the feature model constructed by one reference
-- #1: product line with features with elements, and domain feature model
-- #2: reference with source feature affiliation; target feature affiliation is ignored
-- result: (True,Nothing) if feature model is fully abstract in respect to the reference
--         (False, Just [Feature]) otherwise; the list contains only feature names
-- note: The reason for this method is a smaller reference feature model, it may be checked faster
checkReferenceFullyAbstract :: ProductLine -> ReferenceExpl -> (Bool, Maybe [Feature])
checkReferenceFullyAbstract (fs,dcs) r = hlp (solve1 (problemCnf solverInput))
    where 
        solverInput = propTreeToSolverInput (TNot (TOr (TNot (domainConstraintsToPropTree dcs)) 
                      --(generateRef1DependencyTree fs (Data.Set.findMin(Data.Set.intersection fs 
                      --    (Data.Set.singleton (Feature {featName=fst(fst r),
                      --                                  elements=empty,
                      --                                  references=empty,
                      --                                  contracts=empty})))) (snd$fst r,snd r) )))
                      (generateRefSimpleDependencyTree fs (snd$fst r,snd r) )))
        hlp (Sat ass, stats, Nothing) = 
            (False, Just (enrichFeatures fs 
                (hlp2 (Data.Map.toList (projectCircuitSolution (Sat ass) solverInput)))))
        hlp (Unsat ass, stats, Just resolutionTrace) = 
            (True, Nothing)--Right (genUnsatCore resolutionTrace))
        hlp2 :: [(Feature, Bool)] -> [Feature]
        hlp2 [] = []
        hlp2 ((f,False):fs) = hlp2 fs
        hlp2 ((f,True):fs) = f:(hlp2 fs)

-- checks if the feature model constructed by references implies the given feature model
checkFeatureModelMaximal :: ProductLine -> (Bool, Maybe [Feature]) --(BEnv Feature)
checkFeatureModelMaximal (fs,dcs) = hlp (solve1 (problemCnf solverInput))
    where 
        --additionally: disallow empty product for referential model
        solverInput = propTreeToSolverInput 
                      (TNot (TAnd 
                       (TNot (TAnd 
                        ( (Data.Set.fold (\f -> TAnd (TNot (TLeaf f))) TTrue fs))
                        (generateReferencialFeatureModelTree fs)
                       ))
                       (domainConstraintsToPropTree dcs)
                      ))
        hlp (Sat ass, stats, Nothing) = 
            (False, Just (hlp2 (Data.Map.toList (projectCircuitSolution (Sat ass) solverInput))))
        hlp (Unsat ass, stats, Just resolutionTrace) = 
            (True, Nothing)--Right (genUnsatCore resolutionTrace))
        hlp2 :: [(Feature, Bool)] -> [Feature]
        hlp2 [] = []
        hlp2 ((f,False):fs) = hlp2 fs
        hlp2 ((f,True):fs) = f:(hlp2 fs)

checkFeatureModelMaximalExhaustive :: ProductLine -> (Bool, Set [Feature])
checkFeatureModelMaximalExhaustive pl 
    | Data.Set.size (result) > 0 = (False,result)
    | otherwise                  = (True, result)
    where result = difference (buildAllRefConfigurations pl) (buildAllConfigurations pl)

-- references of #1 which are not supplied by any other out of #1 ; thus essential
-- #1: feature selection to check
-- #2: whole feature set (needed because of)
getUnsuppliedReferences :: [Feature] -> FeatureSet -> [(Feature, Reference)]
getUnsuppliedReferences fs fsAll = 
    [(f,r) | f <- fs, r <- (toList (references f)), 
         Data.Set.isSubsetOf (getFeaturesWithElement fsAll (fst r))
                             (getFeaturesWithElement fsSet (fst r)),
         Data.Set.null (getFeaturesWithElement fsSet (snd r))]
    where fsSet = fromList fs


-- domain constraint data type
-- F1 /\ ... /\ Fk => Nm*Fm \/ ... \/ Nn*Fn    where Ni may negate Fi
data DomainConstraint  = DConstraint [Feature] [FeatureConstraint]
data FeatureConstraint = Must Feature | MustNot Feature
instance Show DomainConstraint where
    show (DConstraint (f:(g:fs)) fss ) = (show f) ++ " /\\ " ++ show (DConstraint (g:fs) fss)
    show (DConstraint (f:[]) fss ) = (show f) ++ " -> " ++ show (DConstraint [] fss)
    show (DConstraint [] ((Must f):(g:fss))) = (show f) ++ " \\/ " ++ show (DConstraint [] (g:fss))
    show (DConstraint [] ((MustNot f):(g:fss))) = "not "++(show f)++" \\/ "++show (DConstraint [] (g:fss))
    show (DConstraint [] ((Must f):[]) ) = (show f)
    show (DConstraint [] ((MustNot f):[]) ) = (show f)

--addDomainConstraint :: ProductLine -> DomainConstraint -> ProductLine
--addDomainConstraint (fs,fm) dc = (fs, CNF {
--    numVars=numVars fm, numClauses=(numClauses fm)+1,
--   clauses=insert (domainConstraintToCNF (toList fs) dc) (clauses fm)} )
    
isDomainConstraintValid :: ProductLine -> DomainConstraint -> Bool
isDomainConstraintValid (fs,fm) (DConstraint pfs cfs) = 
            (isSubsetOf (fromList pfs) fs) && (isSubsetOf (fromList (Prelude.map stripFC cfs)) fs)
    where stripFC (Must f)    = f
          stripFC (MustNot f) = f

domainConstraintsToPropTree :: [DomainConstraint] -> Tree Feature
domainConstraintsToPropTree [] = TTrue
domainConstraintsToPropTree (x:xs) = foldr1 TAnd (Prelude.map domainConstraintToPropTree (x:xs))

domainConstraintToPropTree :: DomainConstraint -> Tree Feature
domainConstraintToPropTree (DConstraint [] []) = TTrue
domainConstraintToPropTree (DConstraint [] ((Must fc):[])) = TLeaf fc
domainConstraintToPropTree (DConstraint [] ((MustNot fc):[])) = TNot (TLeaf fc)
domainConstraintToPropTree (DConstraint [] ((Must fc):(fc2:fcs))) = 
    TOr (TLeaf fc) (domainConstraintToPropTree (DConstraint [] (fc2:fcs)))
domainConstraintToPropTree (DConstraint [] ((MustNot fc):(fc2:fcs))) = 
    TOr (TNot (TLeaf fc)) (domainConstraintToPropTree (DConstraint [] (fc2:fcs)))
domainConstraintToPropTree (DConstraint (f:fs) fcs) = 
    TOr (TNot (TLeaf f)) (domainConstraintToPropTree (DConstraint fs fcs))


domainConstraintToCNF :: [Feature] -> DomainConstraint -> Clause
domainConstraintToCNF fs (DConstraint [] []) = []
domainConstraintToCNF fs (DConstraint [] ((Must f):fcs)) 
    =  (featureToLit fs (featName f)) : (domainConstraintToCNF fs (DConstraint [] fcs))
domainConstraintToCNF fs (DConstraint [] ((MustNot f):fcs))
    = (-featureToLit fs (featName f)) : (domainConstraintToCNF fs (DConstraint [] fcs))
domainConstraintToCNF fs (DConstraint (f:fps) fcs)
    = (-featureToLit fs (featName f)) : (domainConstraintToCNF fs (DConstraint fps fcs))

-- not lifted feature index (first number is 1)
-- features should be atomic
featureToLit :: [Feature] -> String -> Lit
featureToLit fs n = findIndex' fs n 1
    where findIndex' :: [Feature] -> String -> Int -> Lit
          findIndex' [] n m = error "feature does not exist in pl's list of features"
          findIndex' (f:fs) n m | (featName f) == n = L {unLit=m}
                                | otherwise     = findIndex' fs n (m+1)


